@@ -21,86 +21,74 @@ def create_interpolator(method, values, F, M, S, npy=None,degree=-1, log_level=0
         print("Interpolator:",method,F,M,S,"degree=",degree)
     intr = None
     if method == "linear":
-        intr = Multivariate(values,F, M, S,npy, 1, log_level)
+        intr = Multivariate(values,F, M, S,1, log_level)
     elif method == "cubic" and degree == 5: #not really cubic, but quintic
-        intr = Multivariate(values,F, M, S,npy, 5, log_level)
+        intr = Multivariate(values,F, M, S,5, log_level)
     elif method == "cubic":
-        intr = Multivariate(values,F, M, S,npy, 3, log_level)
+        intr = Multivariate(values,F, M, S,3, log_level)
     elif method == "bspline" and degree == 5:
-        intr = Bspline(values,F, M, S,npy, 5, log_level) # I am being cautious about degrees
+        intr = Bspline(values,F, M, S, 5, log_level) # I am being cautious about degrees
     elif method == "bspline":
-        intr = Bspline(values,F, M, S,npy, 3, log_level) 
+        intr = Bspline(values,F, M, S,3, log_level) 
     else: #nearest is default
-        intr = Nearest(values,F, M, S,npy,0,log_level)
+        intr = Nearest(values,F, M, S,0,log_level)
     intr.init()
     return intr
 
 ### Abstract class ############################################################################
 class Interpolator(ABC):
-    def __init__(self, values, F, M, S, npy=None,degree=-1,log_level=0):
-        self.use_jax = False                
-        if npy != None:
-            self._npy = npy
-        else:
-            count = 0
-            self._npy = np.zeros((F,M,S))
-            for k in range(S):
-                for j in range(M):
-                    for i in range(F):
-                        self._npy[i,j,k] = values[count]
-                        count += 1
+    def __init__(self, values, F, M, S, degree=-1,log_level=0):                             
+        self._npy = np.copy(values)
         self._F = F
         self._M = M
         self._S = S
         self.degree = degree
-        self._buffer = 28
+        self._buffer = 12
         self.log_level = log_level         
-        self.h = 0.1 #this is the interval for numerical differentiation
-        
-                
+        self.h = 0.01 #this is the interval for numerical differentiation
+    ############################################################################################################            
     @abstractmethod
     def get_value(self, x, y, z):
         pass
+    
+    @abstractmethod
+    def get_radient(self, x, y, z):
+        pass
+    
+    @abstractmethod
+    def get_laplacian(self, x, y, z):
+        pass
+    ############################################################################################################
 
     # implemented interface that is the same for all abstractions
-    def get_radient(self, x, y, z):                
-        if self.use_jax:
-            grad_radx = grad(self.get_value,argnums=0)(x,y,z)
-            grad_rady = grad(self.get_value,argnums=1)(x,y,z)
-            grad_radz = grad(self.get_value,argnums=2)(x,y,z)
-            dx = grad_radx(x,y,z)
-            dy = grad_rady(x,y,z)
-            dz = grad_radz(x,y,z)
-            radient = (abs(dx) + abs(dy) + abs(dz)) / 3
-            return radient
-        else:
-            val = self.get_value(x, y, z)
-            dx = (self.get_value(x + self.h, y, z) - val) / self.h
-            dy = (self.get_value(x, y + self.h, z) - val) / self.h
-            dz = (self.get_value(x, y, z + self.h) - val) / self.h
-            radient = (abs(dx) + abs(dy) + abs(dz)) / 3
-            return radient
-        
-    def get_laplacian(self, x, y, z):        
+    def get_radient_numerical(self, x, y, z):                        
         val = self.get_value(x, y, z)
-        xx = self.getDxDx(x, y, z, val)
-        yy = self.getDyDy(x, y, z, val)
-        zz = self.getDzDz(x, y, z, val)
+        dx = (self.get_value(x + self.h, y, z) - val) / self.h
+        dy = (self.get_value(x, y + self.h, z) - val) / self.h
+        dz = (self.get_value(x, y, z + self.h) - val) / self.h
+        radient = (abs(dx) + abs(dy) + abs(dz)) / 3
+        return radient
+        
+    def get_laplacian_numerical(self, x, y, z):        
+        val = self.get_value(x, y, z)
+        xx = self.getDxDx_numerical(x, y, z, val)
+        yy = self.getDyDy_numerical(x, y, z, val)
+        zz = self.getDzDz_numerical(x, y, z, val)
         return xx + yy + zz 
         
-    def getDxDx(self, x, y, z, val):        
+    def getDxDx_numerical(self, x, y, z, val):        
         va = self.get_value(x - self.h, y, z)
         vb = self.get_value(x + self.h, y, z)
         dd = (va + vb - 2 * val) / (self.h * self.h)
         return dd
         
-    def getDyDy(self, x, y, z, val):        
+    def getDyDy_numerical(self, x, y, z, val):        
         va = self.get_value(x, y - self.h, z)
         vb = self.get_value(x, y + self.h, z)
         dd = (va + vb - 2 * val) / (self.h * self.h)
         return dd
         
-    def getDzDz(self, x, y, z, val):        
+    def getDzDz_numerical(self, x, y, z, val):        
         va = self.get_value(x, y, z - self.h)
         vb = self.get_value(x, y, z + self.h)
         dd = (va + vb - 2 * val) / (self.h * self.h)
@@ -260,16 +248,20 @@ class Nearest(Interpolator):
         closest_pnt = self.closest(v3.VectorThree(x,y,z))  
         #print(closest_pnt.A, closest_pnt.B,closest_pnt.C)      
         return self.get_fms(closest_pnt.A, closest_pnt.B,closest_pnt.C)    
+    ## implement abstract interface #########################################
+    def get_radient(self, x, y, z):
+        return self.get_radient_numerical(x,y,z)
+    
+    def get_laplacian(self, x, y, z):
+        return self.get_laplacian_numerical(x,y,z)
+    ## iplement abstract interface ###########################################
 ####################################################################################################
 ### Multivariate - Linear and Cubic
 ####################################################################################################
 class Multivariate(Interpolator):                
     def init(self):
-        self.points = self.degree + 1
-        if self.use_jax:
-            self.dimsize = np.power(self.points, 3)
-        else:
-            self.dimsize = math.pow(self.points, 3)
+        self.points = self.degree + 1        
+        self.dimsize = math.pow(self.points, 3)
         if self.degree == 1:
             self.inv = iv1.InvariantVandermonde()
         elif self.degree == 3:
@@ -281,7 +273,30 @@ class Multivariate(Interpolator):
         self._yfloor = -1
         self._zfloor = -1
     
+    ## implement abstract interface #########################################
     def get_value(self, x, y, z):
+        return self.get_value_new(x,y,z)
+
+    def get_radient(self, x, y, z):
+        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
+        dx = self.get_value_multivariate(zn, yn, xn, coeffs,["x"])
+        dy = self.get_value_multivariate(zn, yn, xn, coeffs,["y"])
+        dz = self.get_value_multivariate(zn, yn, xn, coeffs,["z"])
+        return (abs(dx) + abs(dy) + abs(dz))/3
+    
+    def get_laplacian(self, x, y, z):
+        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
+        ddx = self.get_value_multivariate(zn, yn, xn, coeffs,["x","x"])
+        ddy = self.get_value_multivariate(zn, yn, xn, coeffs,["y","y"])
+        ddz = self.get_value_multivariate(zn, yn, xn, coeffs,["z","z"])
+        return ddx+ddy+ddz
+    ## iplement abstract interface ###########################################
+
+    def get_value_new(self, x, y, z):
+        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
+        return self.get_value_multivariate(zn, yn, xn, coeffs)
+
+    def get_value_old(self, x, y, z):
         # The method of linear interpolation is a version of my own method for multivariate fitting, instead of trilinear interpolation
         # NOTE I could extend this to be multivariate not linear but it has no advantage over bspline - and is slower and not as good 
         # Document is here: https://rachelalcraft.github.io/Papers/MultivariateInterpolation/MultivariateInterpolation.pdf                        
@@ -325,18 +340,74 @@ class Multivariate(Interpolator):
         #return self.get_value_multivariate(zn, yn, xn, self.polyCoeffs)
         return self.get_value_multivariate(zn, yn, xn, self.polyCoeffs)
         
-    def get_value_multivariate(self, x, y, z, coeffs):        
+    def get_value_multivariate(self, x, y, z, coeffs,wrt=[]):        
         #This is using a value scheme that makes sens of our new fitted polyCube
         #In a linear case it will be a decimal between 0 and 1          
+        calc_mat = coeffs
+        for partial in wrt:
+            if partial == "x":
+                calc_mat = self.diffWRTx(calc_mat)
+            elif partial == "y":
+                calc_mat = self.diffWRTy(calc_mat)
+            elif partial == "z":
+                calc_mat = self.diffWRTz(calc_mat)
+        
         value = 0
-        ii,jj,kk = coeffs.shape
+        ii,jj,kk = calc_mat.shape
         for i in range(ii):        
             for j in range(jj):            
                 for k in range(kk):                
-                    coeff = coeffs[i, j, k];                    
+                    coeff = calc_mat[i, j, k];                    
                     val = coeff * np.power(z, i) * np.power(y, j) * np.power(x, k)
                     value = value + val                                    
         return value
+
+    def diffWRTx(self,coeffs):        
+        x,y,z = coeffs.shape
+        partialX = np.zeros((x,y,z))
+        for i in range(1, x):
+            for j in range(0,y):
+                for k in range(0,z):
+                    partialX[i-1,j,k] = coeffs[i,j,k]*i
+        return partialX
+
+    def diffWRTy(self,coeffs):        
+        x,y,z = coeffs.shape
+        partialY = np.zeros((x,y,z))
+        for i in range(0, x):
+            for j in range(1,y):
+                for k in range(0,z):
+                    partialY[i,j-1,k] = coeffs[i,j,k]*j
+        return partialY
+    
+    def diffWRTz(self,coeffs):        
+        x,y,z = coeffs.shape
+        partialZ = np.zeros((x,y,z))
+        for i in range(0, x):
+            for j in range(0,y):
+                for k in range(1,z):
+                    partialZ[i,j,k-1] = coeffs[i,j,k]*k
+        return partialZ
+
+    def make_coeffs(self, x, y, z):                                        
+        # 1. Build the points around the centre as a cube - 8 points
+        vals = self.build_cube_around(x, y, z, self.points)
+        #2. Multiply with the precomputed matrix to find the multivariate polynomial
+        ABC = self.mult_vector(self.inv.get_invariant(), vals)
+        # 3. Put the 8 values back into a cube
+        polyCoeffs = np.zeros((self.points, self.points, self.points))
+        pos = 0
+        for i in range(self.points):
+            for j in range(self.points):                
+                for k in range(self.points):                    
+                    polyCoeffs[i, j, k] = ABC[pos]
+                    pos+=1        
+        #4. Adjust the values to be within this cube
+        pstart = (-1 * self.points / 2) + 1        
+        xn = x - np.floor(x) - pstart
+        yn = y - np.floor(y) - pstart
+        zn = z - np.floor(z) - pstart
+        return xn,yn,zn,polyCoeffs
             
 ####################################################################################################
 ### B-Spline
@@ -353,6 +424,14 @@ class Bspline(Interpolator):
         self.mirror = False                
         self.make_periodic_coeffs() #temp make it just ordinary coeffs
             
+    ## implement abstract interface #########################################
+    def get_radient(self, x, y, z):
+        return self.get_radient_numerical(x,y,z)
+    
+    def get_laplacian(self, x, y, z):
+        return self.get_laplacian_numerical(x,y,z)
+    ## iplement abstract interface ###########################################
+
     def make_periodic_coeffs(self):           
         # we make the coefficients matrix a bit bigger than the vaues and have it wrap, and then cut it back down to the values                 
         self._coeffs =self.extend_vals_with_buffer(self._buffer)
@@ -360,33 +439,29 @@ class Bspline(Interpolator):
         tmpM = self._M + (2 * self._buffer)
         tmpS = self._S + (2 * self._buffer)                
         self.create_coeffs(tmpF,tmpM,tmpS)        
-        self._coeffs = self.reduce_vals_with_buffer(self._buffer, self._coeffs,tmpF,tmpM,tmpS)        
-        #self._coeffs = np.copy(self._npy)
-        #self.create_coeffs(self._F,self._M,self._S)
-        
-    def reduce_vals_with_buffer(self,buffer, mynpy,tmpF,tmpM,tmpS):
-        smnpy = np.zeros((self._F,self._M,self._S))
-        for z in range(self._S):            
-            for y in range(self._M):                
-                for x in range(self._F):                    
-                    u_x, u_y, u_z = x+buffer,y+buffer,z+buffer
-                    val = mynpy[u_x,u_y,u_z]
-                    smnpy[x,y,z] = val
+        self._coeffs = self.reduce_vals_with_buffer(self._buffer, self._coeffs)
+                
+    def reduce_vals_with_buffer(self,buffer, mynpy):        
+        xx,yy,zz = mynpy.shape
+        smnpy = mynpy[buffer:xx-buffer,buffer:yy-buffer,buffer:zz-buffer]
         return smnpy
         
     def extend_vals_with_buffer(self,buffer):        
         #// 1. Make a buffer padded values cube for periodic values        
-        xx,yy,zz = self._F, self._M, self._S
-        xx,yy,zz = xx+2*buffer,yy+2*buffer,zz+2*buffer
-        mynpy = np.zeros((xx,yy,zz))
-        for z in range(zz):
-            for y in range(yy):
-                for x in range(xx):
+        xxa,yya,zza = self._F, self._M, self._S        
+        padded = np.pad(self._npy,[(buffer,buffer),(buffer,buffer),(buffer,buffer)])        
+        for z in range(buffer):
+            for y in range(buffer):
+                for x in range(buffer):
                     p = self.get_fms(x-buffer,y-buffer,z-buffer)
-                    mynpy[x,y,z] = p                    
-        return mynpy
-    
-                                              
+                    padded[x,y,z] = p                            
+        for z in range(zza,zza+buffer):
+            for y in range(yya,yya+buffer):
+                for x in range(xxa,xxa+buffer):
+                    p = self.get_fms(x-buffer,y-buffer,z-buffer)                    
+                    padded[x,y,z] = p                                    
+        return padded
+                                                                  
     def create_coeffs(self, thisF, thisM, thisS):
         pole = self.get_pole(self.degree)
         num_poles = len(pole)
@@ -394,23 +469,28 @@ class Bspline(Interpolator):
         #X-wise
         for y in range(thisM):        
             for z in range(thisS):        
-                row = self.get_row3d(y, z, thisF,thisF, thisM, thisS)
-                line = self.convert_to_interp_coeffs(pole, num_poles, thisF, row,self.mirror)
-                self.put_row3d(y, z, line, thisF,thisF, thisM, thisS)
+                #row = self.get_row3d(y, z, thisF,thisF, thisM, thisS)
+                #line = self.convert_to_interp_coeffs(pole, num_poles, thisF, row,self.mirror)
+                #self.put_row3d(y, z, line, thisF,thisF, thisM, thisS)
+                row = self._coeffs[:,y:y+1,z:z+1].reshape([thisF])
+                self.convert_to_interp_coeffs(pole, num_poles, thisF, row,self.mirror)
         #Y-wise
         for x in range(thisF):        
             for z in range(thisS):        
-                row = self.get_col3d(x, z, thisM,thisF, thisM, thisS)
-                line = self.convert_to_interp_coeffs(pole, num_poles, thisM, row,self.mirror)
-                self.put_col3d(x, z, line, thisM,thisF, thisM, thisS)           
+                #row = self.get_col3d(x, z, thisM,thisF, thisM, thisS)
+                #line = self.convert_to_interp_coeffs(pole, num_poles, thisM, row,self.mirror)
+                #self.put_col3d(x, z, line, thisM,thisF, thisM, thisS)           
+                row = self._coeffs[x:x+1,:,z:z+1].reshape([thisM])
+                self.convert_to_interp_coeffs(pole, num_poles, thisM, row,self.mirror)
         #Z-wise
         for x in range(thisF):
             for y in range(thisM):        
-                row = self.get_hole3d(x, y, thisS,thisF, thisM, thisS)
+                #row = self.get_hole3d(x, y, thisS,thisF, thisM, thisS)
+                #line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row,self.mirror)
+                #self.put_hole3d(x, y, line, thisS,thisF, thisM, thisS)
+                row = self._coeffs[x:x+1,y:y+1,:].reshape([thisS])
                 line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row,self.mirror)
-                self.put_hole3d(x, y, line, thisS,thisF, thisM, thisS)
-                                        
-    
+                                            
     def get_pole(self,degree):        
         #Recover the poles from a lookup table #currently only 3 degree, will I want to calculate all the possibilities at the beginnning, 3,5,7,9?
         pole = []
@@ -459,7 +539,7 @@ class Bspline(Interpolator):
     def put_hole3d(self,x,y,bore,length,F,M,S):        
         for z in range(length):
             self.put_coeff(x, y, z, bore[z],F,M,S)
-
+    
     def convert_to_interp_coeffs(self,pole, num_poles, width, row,mirror):        
         #/* special case required by mirror boundaries */
         if (width == 1):         
@@ -516,7 +596,7 @@ class Bspline(Interpolator):
             #/* full loop */
             zn = pole
             iz = 1.0 / pole
-            z2n = np.pow(pole, length - 1)
+            z2n = math.pow(pole, length - 1)
             Sum = vals[0] + z2n * vals[length - 1]
             z2n *= z2n * iz 
             for n in range(1,length - 2):
